#lang racket
(require "run.rkt"
         racket/port
         (for-syntax syntax/parse))

(provide #%module-begin
         (rename-out [pfsh:< <]
                     [pfsh:> >]
                     [pfsh:run #%app]
                     [pfsh:top #%top]
                     [pfsh:define define]
                     [pfsh:datum #%datum]
                     [pfsh:top-interaction #%top-interaction]
                     [pfsh:string-append string-append]))

(module reader syntax/module-reader
  pfsh/main3)

(begin-for-syntax
  (define-syntax-class no-redirect-id
    #:literals (pfsh:< pfsh:>)
    (pattern (~not (~or pfsh:< pfsh:>)))))
    
(define-syntax (pfsh:run stx)
  (syntax-parse stx
    #:literals (pfsh:< pfsh:>)
    [(_ prog arg:no-redirect-id ...
        (~optional (~seq pfsh:< in-stream:expr))
        (~optional (~seq pfsh:> out-stream:id)))
     #:with run-it #`(void (run prog arg ...))
     #:with run-it/in (if (attribute in-stream)
                          #'(with-input-from-string
                              in-stream
                              (lambda ()
                                run-it))
                          #'run-it)
     (if (attribute out-stream)
         #'(define out-stream (with-output-to-string
                                (lambda ()
                                  run-it/in)))
         #'run-it/in)]))

(define-syntax (pfsh:< stx) (raise-syntax-error #f "misplaced" stx))
(define-syntax (pfsh:> stx) (raise-syntax-error #f "misplaced" stx))

(define-syntax (pfsh:top stx)
  (syntax-parse stx
    [(_ . sym:id)
     #`#,(symbol->string (syntax-e #'sym))]))

(define-syntax (pfsh:define stx)
  (syntax-parse stx
    [(_ name:id expr)
     #`(define name expr)]
    [(_ (proc:id arg:id ...) expr)
     #'(begin
         (define (actual-proc arg ...)
           expr)
         (define-syntax (proc stx)
           (syntax-parse stx
             [(_ arg ...) #'(actual-proc arg ...)])))]))

(pfsh:define (pfsh:string-append arg1 arg2)
             (string-append arg1 arg2))

(define-syntax (pfsh:datum stx)
  (syntax-parse stx
    [(_ . s:string) #'(#%datum . s)]
    [(_ . other)
     (raise-syntax-error 'pfsh
                         "only literal strings are allowed"
                         #'other)]))

(define-syntax (pfsh:top-interaction stx)
  (syntax-parse stx
    [(_ . form) #'form]))
