#lang racket
(require "run.rkt"
         racket/port
         (for-syntax syntax/parse))

(provide (rename-out [pfsh:module-begin #%module-begin]
                     [pfsh:< <]
                     [pfsh:> >]
                     [pfsh:run #%app]
                     [pfsh:top #%top]
                     [pfsh:define define]
                     [pfsh:datum #%datum]
                     [pfsh:top-interaction #%top-interaction]
                     [pfsh:string-append string-append]))

(module reader syntax/module-reader
  pfsh/main4)

(define-syntax (pfsh:module-begin stx)
  (syntax-parse stx
    [(_ form ...)
     #`(#%module-begin
        #,@(datum->syntax
            stx
            (let loop ([forms (syntax->list #'(form ...))])
              (cond
                [(null? forms) null]
                [(pair? (syntax-e (car forms)))
                 (cons (car forms) (loop (cdr forms)))]
                [else
                 (define line (syntax-line (car forms)))
                 (let line-loop ([forms (cdr forms)] [accum (list (car forms))])
                   (cond
                     [(null? forms) (list (reverse accum))]
                     [(eqv? line (syntax-line (car forms)))
                      (line-loop (cdr forms) (cons (car forms) accum))]
                     [else (cons (reverse accum) (loop forms))]))]))))]))

(begin-for-syntax
  (define-syntax-class no-redirect-id
    #:literals (pfsh:< pfsh:>)
    (pattern (~not (~or pfsh:< pfsh:>)))))
    
(define-syntax (pfsh:run stx)
  (syntax-parse stx
    #:literals (pfsh:< pfsh:>)
    [(_ prog arg:no-redirect-id ...
        (~optional (~seq pfsh:< in-stream:expr))
        (~optional (~seq pfsh:> out-stream:id)))
     #:with run-it #`(void (run prog arg ...))
     #:with run-it/in (if (attribute in-stream)
                          #'(with-input-from-string
                              in-stream
                              (lambda ()
                                run-it))
                          #'run-it)
     (if (attribute out-stream)
         #'(define out-stream (with-output-to-string
                                (lambda ()
                                  run-it/in)))
         #'run-it/in)]))

(define-syntax (pfsh:< stx) (raise-syntax-error #f "misplaced" stx))
(define-syntax (pfsh:> stx) (raise-syntax-error #f "misplaced" stx))

(define-syntax (pfsh:top stx)
  (syntax-parse stx
    [(_ . sym:id)
     #`#,(symbol->string (syntax-e #'sym))]))

(define-syntax (pfsh:define stx)
  (syntax-parse stx
    [(_ name:id expr)
     #`(define name expr)]
    [(_ (proc:id arg:id ...) expr)
     #'(begin
         (define (actual-proc arg ...)
           expr)
         (define-syntax (proc stx)
           (syntax-parse stx
             [(_ arg ...) #'(actual-proc arg ...)])))]))

(pfsh:define (pfsh:string-append arg1 arg2)
             (string-append arg1 arg2))

(define-syntax (pfsh:datum stx)
  (syntax-parse stx
    [(_ . s:string) #'(#%datum . s)]
    [(_ . other)
     (raise-syntax-error 'pfsh
                         "only literal strings are allowed"
                         #'other)]))

(define-syntax (pfsh:top-interaction stx)
  (syntax-parse stx
    [(_ . form) #'form]))
